#include <stdio.h>
#include <stdlib.h>
#include <SDL/SDL.h>
#include "types.h"
#include "defines.h"

/* defines */

/* None */

/* external variables */

extern fixed sintable[ANGLES+ANGLES/4];
extern S16 pixelangle[SCREEN_WIDTH];
extern finetangent[FINEANGLES/4];
extern fixed finecos[SCREEN_WIDTH];
extern pl_point_type player;
extern U8 Map[MAP_HEIGHT][MAP_WIDTH];

/* global variables */

/* None */

/* local variables */

static S32          xintercept,yintercept;
static fixed        viewsin,viewcos;
static point_type   gridstep;
static fixed        *costable = sintable+(ANGLES/4);
static SDL_Surface* scr;

/* local prototypes */

static void CalcHeight(U16 *SliceHeight);
static void DrawSlice(S16 x, color_type Color, U16 SliceHeight);
static void verLine(S32 x, S32 y1, S32 y2, color_type color);
static void redraw(void);
static void cls(void);
static void HitHorizWall(S16 pixx, color_type Color);
static void HitVertWall(S16 pixx, color_type Color);

/* 
================================================================
=
= Function: Draw3D
=
= Description:
=
= Raycast simple 3d level onto screen
=
================================================================
*/
void Draw3D(void)
{
        U32         xpartialleft;
        U32         xpartialright;
        U32         ypartialup;
        U32         ypartialdown;
        S32         xstep,ystep;
        U16         xspot,yspot;
        U32         xpartial,ypartial;
        U16         xgrid,ygrid;
        U16         gridhit;
        S16         viewangle,midangle;
        color_type  Color;
        S16         pixx;
        fixed       viewx;
        fixed       viewy;

        viewangle = player.angle;
        midangle = viewangle*(FINEANGLES/ANGLES);
        viewsin = sintable[viewangle];
        viewcos = costable[viewangle];

        /* distance to left side of grid */
        xpartialleft = player.x & (BIN16-1);
        /* distance to right side of grid */
        xpartialright = BIN16 - xpartialleft;
        /* distance to top of grid */
        ypartialup = player.y & (BIN16-1);
        /* distance to bottom of grid */
        ypartialdown = BIN16 - ypartialup;

        /* calculate each ray within the scene */
        for(pixx = 0 ; pixx < SCREEN_WIDTH ; pixx++)
        {
            S16 angl = midangle + pixelangle[pixx];

            if(angl<0)
            {
                angl+=FINEANGLES;
            }
            if(angl>=3600)
            {
                angl-=FINEANGLES;
            }

            if(angl<900)
            {
                /* next grid location from present location */
                gridstep.x = 1;
                gridstep.y = -1;
                /* x or y distance we are going to add for this grid step */
                xstep = finetangent[900-1-angl];
                ystep = -finetangent[angl];
                /* distance to grid sides for this angle */
                xpartial = xpartialright;
                ypartial = ypartialup;
            }
            else if(angl<1800)
            {
                gridstep.x = -1;
                gridstep.y = -1;
                xstep = -finetangent[angl-900];
                ystep = -finetangent[1800-1-angl];
                xpartial = xpartialleft;
                ypartial = ypartialup;
            }
            else if(angl<2700)
            {
                gridstep.x = -1;
                gridstep.y = 1;
                xstep = -finetangent[2700-1-angl];
                ystep = finetangent[angl-1800];
                xpartial = xpartialleft;
                ypartial = ypartialdown;
            }
            else if(angl<3600)
            {
                gridstep.x = 1;
                gridstep.y = 1;
                xstep = finetangent[angl-2700];
                ystep = finetangent[3600-1-angl];
                xpartial = xpartialright;
                ypartial = ypartialdown;
            }

            /* work out first y intercept location of the vertial ray */
            yintercept = FixedMul(ystep,xpartial) + player.y;
            /* next x grid location vertical ray will be in */
            xgrid = (player.x >> MAP_TO_GRID_SHIFT) + gridstep.x;
            /* next grid location vertical ray will be in */
            xspot = (U16)(xgrid + (((U32)yintercept >> MAP_TO_GRID_SHIFT) << MAP_WIDTH_SHIFT));

            /* work out first x intercept location of the horizontal ray */
            xintercept = FixedMul(xstep,ypartial) + player.x;
            /* next y grid location horizontal ray will be in */
            ygrid = (player.y >> MAP_TO_GRID_SHIFT) + gridstep.y;
            /* next grid location horizontal ray will be in */
            yspot = (U16)(((U32)xintercept >> MAP_TO_GRID_SHIFT) + (ygrid << MAP_WIDTH_SHIFT));

            /* vertical ray casting */
            do
            {
                if(gridstep.y == -1 && (yintercept >> 16) <= ygrid) goto horizentry;
                if(gridstep.y == 1 && (yintercept >> 16) >= ygrid) goto horizentry;
vertentry:
                /* get value of grid that ray has been cast at */
                gridhit = ((U8 *)Map)[xspot];

                /* does location of ray hold any wall value? */
                if(gridhit != 0)
                {
                    /* choose wall colour */
                    switch(gridhit)
                    {
                        case 1:     /* red */
                            Color.r = 255;
                            Color.g = 0;
                            Color.b = 0;
                        break;

                        case 2:     /* green */
                            Color.r = 0;
                            Color.g = 255;
                            Color.b = 0;
                        break;

                        case 3:     /* blue */
                            Color.r = 0;
                            Color.g = 0;
                            Color.b = 255;
                        break;
                    }
                    xintercept = xgrid << MAP_TO_GRID_SHIFT;
                    ygrid = (S16) (yintercept >> MAP_TO_GRID_SHIFT);
                    HitVertWall(pixx, Color);
                    break;
                }

                /* else move ray to next grid coord's */
                xgrid += gridstep.x;
                yintercept += ystep;
                xspot = (U16)(xgrid + (((U32)yintercept >> 16) << MAP_WIDTH_SHIFT));
            }while(1);

            /* cast next ray */
            continue;

            /* horizontal ray casting */
            do
            {
                if(gridstep.x == -1 && (xintercept >> 16) <= xgrid) goto vertentry;
                if(gridstep.x == 1 && (xintercept >> 16) >= xgrid) goto vertentry;
horizentry:
                /* get value of grid that ray has been cast at */
                gridhit=((U8 *)Map)[yspot];

                /* does location of ray hold any wall value? */
                if(gridhit != 0)
                {
                     /* choose wall colour */
                    switch(gridhit)
                    {
                        case 1:     /* red */
                            Color.r = 255;
                            Color.g = 0;
                            Color.b = 0;
                        break;

                        case 2:     /* green */
                            Color.r = 0;
                            Color.g = 255;
                            Color.b = 0;
                        break;

                        case 3:     /* blue */
                            Color.r = 0;
                            Color.g = 0;
                            Color.b = 255;
                        break;
                    }
                    yintercept = ygrid << MAP_TO_GRID_SHIFT;
                    xgrid = (S16) (xintercept >> MAP_TO_GRID_SHIFT);
                    HitHorizWall(pixx, Color);
                    break;
                }

                /* else move ray to next grid coord's */
                ygrid += gridstep.y;
                xintercept += xstep;
                yspot = (U16)(((U32)xintercept >> MAP_TO_GRID_SHIFT) + (ygrid << MAP_WIDTH_SHIFT));
            }while(1);

        }/* end of for loop */

    redraw();
    cls();
}

/* 
================================================================
=
= Function: HitHorizWall
=
= Description:
=
= Ray hit wall on Horizontal line of grid so calculate distance 
= and display wall height.
=
================================================================
*/
static void HitHorizWall(S16 pixx, color_type Color)
{
    U16 SliceHeight;

    if(gridstep.y == -1)
    {
        yintercept += GRID_SIZE;
    }

    CalcHeight(&SliceHeight);
    DrawSlice(pixx, Color, SliceHeight);
}

/* 
================================================================
=
= Function: HitVertWall
=
= Description:
=
= Ray hit wall on vertical line of grid so calculate distance 
= and display wall height.
=
================================================================
*/
static void HitVertWall(S16 pixx, color_type Color)
{
    U16 SliceHeight;

    if(gridstep.x == -1)
    {
        xintercept += GRID_SIZE;
    }

    CalcHeight(&SliceHeight);
    DrawSlice(pixx, Color, SliceHeight);
}

/* 
================================================================
=
= Function: CalcHeight
=
= Description:
=
= Calculates total distance that the ray traveled in x / y 
= directions then convert these distances into a x vector
= and a y vector. Then it subracts one vector from the other to 
= get total vector distance travelled by ray. It then uses
= this distance (z) to work out the correct wall height to 
= display. 
=
================================================================
*/
static void CalcHeight(U16 *SliceHeight)
{
    fixed z = FixedMul(xintercept - player.x, viewcos)
        - FixedMul(yintercept - player.y, viewsin);

    *SliceHeight = PROJECTION_RATIO_SCALED / z ;
}

/* 
================================================================
=
= Function: DrawSlice
=
= Description:
=
= Draws the wall height onto the screen centred around the
= screens midpoint.
=
================================================================
*/
static void DrawSlice(S16 x, color_type Color, U16 SliceHeight)
{
    S32 drawStart;
    S32 drawEnd;

    /* clip height to screen height if needed */
    if(SliceHeight > SCREEN_HEIGHT)
    {
        SliceHeight = SCREEN_HEIGHT;
    }

    drawStart = SCREEN_CENTRE - (SliceHeight / 2);
    drawEnd = drawStart + SliceHeight;
  
    /* draw the pixels of the stripe as a vertical line */
    verLine((S32)x, drawStart, drawEnd, Color);
}

/* 
================================================================
=
= Function: 
=
= Description:
=
= Fast vertical line from (x,y1) to (x,y2), with rgb color 
=
================================================================
*/
static void verLine(S32 x, S32 y1, S32 y2, color_type color)
{
  S32 y;

  if(y2 < y1) {y1 += y2; y2 = y1 - y2; y1 -= y2;}       /* swap y1 and y2 */
  if(y2 < 0 || y1 >= 200  || x < 0 || x >= 320) return; /* no single point of the line is on screen */
  if(y1 < 0) y1 = 0;                                    /* clip */
  if(y2 >= 320) y2 = 200 - 1;                           /* clip */

  U32 colorSDL = SDL_MapRGB(scr->format, color.r, color.g, color.b);
  U32* bufp;

  bufp = (U32*)scr->pixels + y1 * scr->pitch / 4 + x;

  for(y = y1; y <= y2; y++)
  {
     *bufp = colorSDL;
     bufp += scr->pitch / 4;
  }
}

/* 
================================================================
=
= Function: redraw
=
= Description:
=
= Updates the screen.  Has to be called to view new pixels, 
= but use only after drawing the whole screen because it's slow.
=
================================================================
*/
static void redraw(void)
{
  SDL_UpdateRect(scr, 0, 0, 0, 0);
}

/* 
================================================================
=
= Function: cls
=
= Description:
=
= Clears the screen to black
=
================================================================
*/
static void cls(void)
{
  SDL_FillRect(scr, NULL, 0);
}

/* 
================================================================
=
= Function: InitSDLScreenBuf
=
= Description:
=
= Try and get a handle to the screen buffer
=
================================================================
*/
U8 InitSDLScreenBuf(void)
{
    S32 Width = SCREEN_WIDTH;
    S32 Height = SCREEN_HEIGHT;
    S32 ColorDepth = 32;

     /* use SDL lib to get access to screen buffer */
    if(SDL_Init(SDL_INIT_EVERYTHING) < 0)
    {
        printf("Unable to init SDL: %s\n", SDL_GetError());
        SDL_Quit();
        return(1);
    }

    scr = SDL_SetVideoMode(Width, Height, ColorDepth, SDL_HWSURFACE | SDL_HWPALETTE);

    if(scr == NULL)
    {
        printf("Unable to set video: %s\n", SDL_GetError());
        SDL_Quit();
        return(1);
    }

    return(0);
}
