#include <stdio.h>
#include <stdlib.h>
#include "types.h"
#include "defines.h"

/* defines */

#define BASEMOVE        35
#define ANGLESCALE      20
#define MOVESCALE       150l
#define PLAYERSIZE      0x5800l     /* player radius */

/* external variables */

extern fixed sintable[ANGLES+ANGLES/4];
extern U8 Map[MAP_HEIGHT][MAP_WIDTH];

/* global variables */

/* define players start position x = 8.5 y = 8.5 angle = 90 */
pl_point_type player = {(8.5 * GRID_SIZE),(8.5 * GRID_SIZE),90};

/* local variables */

static S16   anglefrac = 0;
static fixed *costable = sintable+(ANGLES/4);

/* local prototypes */

static void Thrust(S32 angle, S32 speed);
static void ClipMove(S32 xmove, S32 ymove);
static U8 TryMove(void);

/* 
================================================================
=
= Function: MovePlayer
=
= Description:
=
= check to see if user has pressed the arrow keys if so it
= moves the player in the direction the user wants.
=
================================================================
*/
void MovePlayer(input_type *Controls)
{
    S32    angle; 
    S32    angleunits;
  
    /* rotate to the right */
    if(Controls->Right != 0)
    {
        anglefrac += BASEMOVE;
        angleunits = anglefrac / ANGLESCALE;
        anglefrac -= angleunits * ANGLESCALE;
        player.angle -= angleunits;               
        
        if(player.angle < 0)
        {
            player.angle += ANGLES;
        }
       
    }

    /* rotate to the left */
    if(Controls->Left != 0)
    {
        anglefrac -= BASEMOVE;
        angleunits = anglefrac / ANGLESCALE;
        anglefrac -= angleunits * ANGLESCALE;
        player.angle -= angleunits;               
        
        if(player.angle >= ANGLES)
        {
            player.angle -= ANGLES;
        }
    }
    
    /* move forward */
    if(Controls->Forward != 0)
    {
        Thrust(player.angle,BASEMOVE*MOVESCALE);                
    }
    
    /* move back */
    if(Controls->Back != 0)
    {   
        /* flip angle so new x , y coord's are behind player */
        angle = player.angle + ANGLES/2;
        
        if (angle >= ANGLES)
        {
            angle -= ANGLES;
        }
        
        Thrust(angle,BASEMOVE*MOVESCALE);                
    }
    
}

/* 
================================================================
=
= Function: Thrust
=
= Description:
=
= calculates new speed and direction of player 
=
================================================================ 
*/
static void Thrust(S32 angle, S32 speed)
{
    S32 xmove;
    S32 ymove;
    
    /* calculate new x and y positions of player*/
    xmove = FixedMul(speed,costable[angle]);
    ymove = -FixedMul(speed,sintable[angle]);
    
    /* check to see if new positions are valid on map*/
    ClipMove(xmove,ymove);    
}

/* 
================================================================
=
= Function: ClipMove
=
= Description:
=
= move player in the direction the user wants if there are no 
= walls present at the new corrd's
=
================================================================ 
*/
static void ClipMove(S32 xmove, S32 ymove)
{
    S32    basex;
    S32    basey;

    /* store current position */
    basex = player.x;
    basey = player.y;
    
    /* add new position to current one */
    player.x = basex + xmove;
    player.y = basey + ymove;
    
    /* check if move is ok */
    if(TryMove() == 1)
    {
        return;
    }
    
    /* try moving in the x direction only */
    player.x = basex + xmove;
    player.y = basey;
    
    if(TryMove ())
    {
        return;
    }
    
    /* try moving in the y direction only */
    player.x = basex;
    player.y = basey+ymove;
    
    if (TryMove())
    {
        return;
    }
    
    /* no luck stay in current location */
    player.x = basex;
    player.y = basey;
}

/* 
================================================================
=
= Function: TryMove
=
= Description:
=
= returns true if move ok
=
================================================================ 
*/
static U8 TryMove(void)
{
    S32    xl;
    S32    yl;
    S32    xh;
    S32    yh;
    S32    x;
    S32    y;
    U8     check;


    /* work out players max x y size in grid coord's*/
    xl = (player.x - PLAYERSIZE) >> MAP_TO_GRID_SHIFT;
    yl = (player.y - PLAYERSIZE) >> MAP_TO_GRID_SHIFT;

    xh = (player.x + PLAYERSIZE) >> MAP_TO_GRID_SHIFT;
    yh = (player.y + PLAYERSIZE) >> MAP_TO_GRID_SHIFT;
    
    /* check for solid walls */
    for(y = yl ; y <= yh; y++)
    {
        for(x = xl; x <= xh; x++)
        {
            check = Map[y][x];
            
            if(check != 0)
            {
                return 0;
            }
        }
    }
    
    return 1;

}
